import numpy as np
from scipy.special import binom

from .grassmannian import *

def spin_basis(j, m):
    return np.eye(int(2*j+1))[int(j-m)]

def sigma_plus(j):
    return np.array([[np.sqrt(j*(j+1) - m*n) \
                          if m == n+1 else 0 \
                              for n in np.arange(j, -j-1, -1)]\
                                 for m in np.arange(j, -j-1, -1)])

def sigma_minus(j):
    return sigma_plus(j).conj().T

def sigma_z(j):
    return np.diag(np.array([m for m in np.arange(j, -j-1, -1)]))

def sigma_y(j):
    return (sigma_plus(j) - sigma_minus(j))/(2j)

def sigma_x(j):
    return (sigma_plus(j) + sigma_minus(j))/2

###

def majorana_polynomial(ket):
    j = (len(ket)-1)/2
    return np.polynomial.Polynomial(\
                    [(-1)**(j-m)*\
                     np.sqrt(binom(2*j, j-m))*\
                     spin_basis(j, m).conj() @ ket\
                         for m in np.arange(-j, j+1)])

def projective_roots(n, poly):
    r = poly.roots()
    return np.concatenate([r, np.repeat(np.inf, n-len(r))])

def majorana_roots(ket):
    return projective_roots(len(ket)-1, majorana_polynomial(ket))

def plane_to_sphere(z):
    if z == np.inf:
        return np.array([0,0,-1])
    else:
        return np.array([2*z.real, 2*z.imag, 1-z.real**2 - z.imag**2])/(1+z.real**2+z.imag**2)
    
def projective_stars(n, poly):
    return np.array([plane_to_sphere(r) for r in projective_roots(n, poly)])

def majorana_stars(ket):
    return projective_stars(len(ket)-1, majorana_polynomial(ket))

###

def spin_operator_to_plucker(O, k):
    n = len(O)
    P = plucker_basis(k, n)
    return P @ sum([reduce(np.kron, \
                               [O if i == j else np.eye(n) \
                                   for j in range(k)])\
                                       for i in range(k)]) @ P.T/k

def lower_until_zero(j, k, highest=None):
    N = int(binom(int(2*j+1), k))
    highest = highest if type(highest) == np.ndarray else basis(N, 0)
    rungs = [highest]
    L = spin_operator_to_plucker(sigma_minus(j), k)
    while not np.allclose(rungs[-1], np.zeros(N)):
        rungs[-1] = rungs[-1]/np.linalg.norm(rungs[-1])
        rungs.append(L @ rungs[-1])
    return np.array(rungs[:-1])

###

# From qutip
def _factorial_prod(N, arr):
    arr[:int(N)] += 1

def _factorial_div(N, arr):
    arr[:int(N)] -= 1

def _to_long(arr):
    prod = 1
    for i, v in enumerate(arr):
        prod *= (i+1)**int(v)
    return prod

def clebsch(j1, j2, j3, m1, m2, m3):
    if m3 != m1 + m2:
        return 0
    vmin = int(np.max([-j1 + j2 + m3, -j1 + m1, 0]))
    vmax = int(np.min([j2 + j3 + m1, j3 - j1 + j2, j3 + m3]))

    c_factor = np.zeros((int(j1 + j2 + j3 + 1)), np.int32)
    _factorial_prod(j3 + j1 - j2, c_factor)
    _factorial_prod(j3 - j1 + j2, c_factor)
    _factorial_prod(j1 + j2 - j3, c_factor)
    _factorial_prod(j3 + m3, c_factor)
    _factorial_prod(j3 - m3, c_factor)
    _factorial_div(j1 + j2 + j3 + 1, c_factor)
    _factorial_div(j1 - m1, c_factor)
    _factorial_div(j1 + m1, c_factor)
    _factorial_div(j2 - m2, c_factor)
    _factorial_div(j2 + m2, c_factor)
    C = np.sqrt((2.0 * j3 + 1.0)*_to_long(c_factor))

    s_factors = np.zeros(((vmax + 1 - vmin), (int(j1 + j2 + j3))), np.int32)
    sign = (-1) ** (vmin + j2 + m2)
    for i,v in enumerate(range(vmin, vmax + 1)):
        factor = s_factors[i,:]
        _factorial_prod(j2 + j3 + m1 - v, factor)
        _factorial_prod(j1 - m1 + v, factor)
        _factorial_div(j3 - j1 + j2 - v, factor)
        _factorial_div(j3 + m3 - v, factor)
        _factorial_div(v + j1 - j2 - m3, factor)
        _factorial_div(v, factor)
    common_denominator = -np.min(s_factors, axis=0)
    numerators = s_factors + common_denominator
    S = sum([(-1)**i * _to_long(vec) for i,vec in enumerate(numerators)]) * \
        sign / _to_long(common_denominator)
    return C * S

spherical_tensor_cache = {}

def spherical_tensor(j, l, m):
    if (j, l, m) in spherical_tensor_cache:
        return spherical_tensor_cache[(j, l, m)]
    T = np.array([[(-1.)**(j-m1-m)*clebsch(j, j, l, m1, -m2, m) \
                   for m1 in np.arange(j, -j-1, -1)]\
                          for m2 in np.arange(j, -j-1, -1)])
    spherical_tensor_cache[(j, l, m)] = T
    return T

def multipole_expansion(O):
    j = (len(O)-1)/2
    return np.array([[(spherical_tensor(j, l, m).conj().T @ O).trace() for m in np.arange(-l, l+1)] for l in np.arange(0, 2*j+1)])

###

def expect_xyz(ket):
    j = (len(ket)-1)/2
    return np.array([ket.conj() @ sigma_x(j) @ ket,\
                     ket.conj() @ sigma_y(j) @ ket,\
                     ket.conj() @ sigma_z(j) @ ket]).real

def rotate(j, theta, phi):
    return sc.linalg.expm((theta/2) * \
                          (np.exp(-1j*phi)*sigma_plus(j) -\
                           np.exp(1j*phi)*sigma_minus(j)))

def cartesian_to_spherical(xyz):
    x, y, z = xyz
    r = np.linalg.norm(xyz)
    return np.array([r,np.arccos(z/r),np.arctan2(y,x)])

###

def first_nonzero_in_multipole_expansion(ket):
    j = (len(ket)-1)/2
    MP = multipole_expansion(np.outer(ket, ket.conj()))
    for l in np.arange(0, 2*j+1):
        for m in np.arange(-l, l+1):
            c = MP[int(l)][int(m+l)]
            if not np.isclose(c,0) and m != 0:
                return [c, l, m]
            
def spectator_component(ket):
    if len(ket) == 1:
        return ket[0]
    j = (len(ket)-1)/2
    r, theta, phi = cartesian_to_spherical(expect_xyz(ket))
    ket_ = rotate(j, theta, phi) @ ket
    c, l, m = first_nonzero_in_multipole_expansion(ket_)
    ket__ = sc.linalg.expm(-1j*np.angle(c)*sigma_z(j)/m) @ ket_
    c2 = ket__[np.where(np.logical_not(np.isclose(ket__, 0)))[0][0]]
    return np.sqrt(ket.conj() @ ket)*np.exp(1j*np.angle(c2))

###

def check_1anticoherence(G):
    k, n = G.shape
    j = (n-1)/2
    return np.allclose(np.array([G @ sigma_x(j) @ G.conj().T,\
                                 G @ sigma_y(j) @ G.conj().T,\
                                 G @ sigma_z(j) @ G.conj().T]),0)
def qmatrix(f, t):
    k = len(f(0))
    return np.array([[f(0)[i].conj() @ f(t)[j] for j in range(k)] for i in range(k)])

def holonomy(f, t):
    U, D, V = np.linalg.svd(qmatrix(f, t))
    return U @ V

def rot_holonomy(G, R):
    k = G.shape[0]
    G_ = G @ R
    Q = np.array([[G[i].conj() @ G_[j] for j in range(k)] for i in range(k)])
    U, D, V = np.linalg.svd(Q)
    return U @ V

###

def sign(lst):
    parity = 1
    for i in range(0,len(lst)-1):
        if lst[i] != i:
            parity *= -1
            mn = min(range(i,len(lst)), key=lst.__getitem__)
            lst[i],lst[mn] = lst[mn],lst[i]
    return parity    

def principle_constellation_polynomial(G):
    k, n = G.shape
    P = [majorana_polynomial(g) for g in G]
    W = np.array([[P[i].deriv(m=j) for j in range(k)] for i in range(k)])
    return sum([sign(list(perm))*np.prod([W[i,p] for i, p in enumerate(perm)]) for perm in permutations(list(range(k)))]) 

def number_of_principal_stars(j, k):
    return 2*np.sum(np.arange(j-k+1, j+1))

###

def spin_coherent_state(j, z):
    if z == np.inf:
        return spin_coherent_state(j, 0)[::-1]
    return (1+z*z.conjugate())**(-j) * \
            sum([np.sqrt(binom(2*j, j-m))*z**(j-m)*spin_basis(j,m)\
                     for m in np.arange(-j, j+1)])

##

def integer(x):
    return np.equal(np.mod(x, 1), 0)

def spin_coherent_representation(ket):
    j = (len(ket)-1)/2
    m = majorana_polynomial(ket)
    def scs_rep(z):
        if z == np.inf:
            return ket[0]
        elif np.isclose(z, 0):
            return ket[-1]
        return ((z.conjugate()/z)/(1+z*z.conjugate()))**j * m(z)
    return scs_rep

###

def coherent_plane(j, k):
    d = int(2*j+1)
    return lambda z: np.eye(d)[:k,:]@sc.linalg.expm(-z*sigma_plus(j)) 

###

import jax
import jax.numpy as jp
from jax.config import config
config.update("jax_enable_x64", True)

def find_1anticoherent_subspace(j, k, max_iter=1000):
    d = int(2*j + 1)
    X, Y, Z = sigma_x(j), sigma_y(j), sigma_z(j)

    @jax.jit
    def one_anticoherence(V):
        R = (V[0:d*k] + 1j*V[d*k:]).reshape(k, d)
        return (jp.linalg.norm(R @ X @ R.conj().T) + \
                jp.linalg.norm(R @ Y @ R.conj().T) + \
                jp.linalg.norm(R @ Z @ R.conj().T)).real

    @jax.jit
    def orthogonality(V):
        R = (V[0:d*k] + 1j*V[d*k:]).reshape(k, d)
        return jp.linalg.norm((R @ R.conj().T) - jp.eye(k)).real
    
    for t in range(max_iter):
        try:
            V = np.random.randn(2*d*k)
            result = sc.optimize.minimize(one_anticoherence, V,\
                                          jac=jax.jit(jax.jacrev(one_anticoherence)),\
                                          tol=1e-23,\
                                          constraints=[{"type": "eq",\
                                                        "fun": orthogonality,\
                                                        "jac": jax.jit(jax.jacrev(orthogonality))}],\
                                          options={"disp": True,\
                                                   "maxiter": 5000},
                                          method="trust-constr")
            R = (result.x[0:d*k] + 1j*result.x[d*k:]).reshape(k, d)
            return R
        except:
            continue

##




import numpy as np
import scipy as sc
import scipy.linalg
from scipy.stats import ortho_group
from scipy.stats import unitary_group

import jax
import jax.numpy as jp
from jax.config import config
config.update('jax_platform_name', 'cpu')
config.update("jax_enable_x64", True)

from .utils import *
from .povm_utils import *

def rand_ginibre(*shape, field="complex"):
    r"""
    Random sample from Ginibre ensemble.
    """
    if field == "complex":
        return np.random.randn(*shape) + 1j*np.random.randn(*shape)
    elif field == "real":
        return np.random.randn(*shape)

def rand_ket(d, field="complex"):
    r"""
    Random ket.
    """
    return normalize(rand_ginibre(d, 1, field=field))

def rand_ket_hs(d, field="complex"):
    if field == "complex": 
        return (unitary_group.rvs(d) @ np.eye(d)[0]).reshape(d, 1)
    elif field == "real":
        return ortho_group.rvs(d) @ np.eye(d)[0].reshape(d, 1)
        
def rand_herm(d, field="complex"):
    r"""
    Random Hermitian (symmetric) matrix.
    """
    X = rand_ginibre(d, d, field=field)
    return (X + X.conj().T)/2

def rand_dm(d, r=None, field="complex"):
    r"""
    Random density matrix.
    """
    r = r if type(r) != type(None) else d
    X = rand_ginibre(d, r, field=field)
    rho = X @ X.conj().T
    return rho/rho.trace()

def rand_unitary(d, field="complex"):
    """
    Random unitary (orthogonal) matrix.
    """
    if field == "complex":
        return unitary_group.rvs(d)
    elif field == "real":
        return ortho_group.rvs(d)

def rand_povm(d, n=None, r=None, field="complex"):
    r"""
    Generates a Haar distributed random POVM for a Hilbert space of dimension $d$, 
    with $n$ elements, and with rank $m$.

    $m$ must satisfy $d \leq mn$, and defaults to $m=1$, giving rank-1 POVM elements.

    $n$ defaults to $d^2$ if complex, $\frac{d(d+1)}{2}$ if real.
    """
    n = n if type(n) != type(None) else state_space_dimension(d, field)
    r = r if type(r) != type(None) else d
    if field == "complex":
        povm = np.zeros((n, d, d), dtype=np.complex128) 
        S = np.zeros(d, dtype=np.complex128) 
    elif field == "real":
        povm = np.zeros((n, d, d))
        S = np.zeros(d)

    for i in range(n):
        Xi = rand_ginibre(d, r, field=field)
        Wi = Xi @ Xi.conjugate().T
        povm[i, :, :] = Wi
        S = S + Wi
    S = sc.linalg.fractional_matrix_power(S, -1/2)
    for i in range(n):
        Wi = np.squeeze(povm[i, :, :])
        povm[i, :, :] = S @ Wi @ S
    return povm

def rand_effect(d, n=None, r=None, field="complex"):
    r"""
    Generates a Haar distributed random POVM effect of Hilbert space dimension $d$, 
    as if it were part of a POVM of $n$ elements with rank $m$. 
    """
    n = n if type(n) != type(None) else state_space_dimension(d, field)
    r = r if type(r) != type(None) else d

    X = rand_ginibre(d, r, field=field)
    W = X @ X.conjugate().T
    Y = rand_ginibre(d, (n-1)*r, field=field)
    S = W + Y @ Y.conjugate().T
    S = sc.linalg.fractional_matrix_power(S, -1/2)
    return S @ W @ S.conjugate().T

def rand_ftf(d, n=None, field="complex"):
    """
    Random tight frame.
    """
    n = n if type(n) != type(None) else state_space_dimension(d, field)
    if field == "complex":
        R = np.random.randn(d, n) + 1j*np.random.randn(d, n) 
    elif field == "real":
        R = np.random.randn(d, n)
    return tighten(R)

def rand_funtf(d, n=None, field="complex", rtol=1e-15, atol=1e-15):
    r"""
    Random finite unit norm tight frame.
    """
    n = n if type(n) != type(None) else state_space_dimension(d, field)
    if field == "complex":
        R = np.random.randn(d, n) + 1j*np.random.randn(d, n) 
    elif field == "real":
        R = np.random.randn(d, n)
    while not (np.allclose(R @ R.conj().T, (n/d)*np.eye(d), rtol=rtol, atol=atol) and\
               np.allclose(np.linalg.norm(R, axis=0), np.ones(n), rtol=rtol, atol=atol)):
        R = sc.linalg.polar(R)[0]
        R = np.array([state/np.linalg.norm(state) for state in R.T]).T
    return sc.linalg.polar(R)[0]

def rand_kraus(d, n, field="complex"):
    r"""
    Random Kraus operators.
    """
    G = [rand_ginibre(d, d, field=field) for i in range(n)]
    H = sum([g.conj().T @ g for g in G])
    S = sc.linalg.fractional_matrix_power(H, -1/2)
    return np.array([g @ S for g in G])

def rand_probs(n, m):
    r"""m random probability vectors of length n."""
    return np.random.dirichlet((1,)*n, size=m).T

def rand_probs_table(m, n, r):
    if r < 2:
        raise Exception("r must be > 1")
    P = np.vstack([np.ones((1,n)), np.random.uniform(low=0, high=1, size=(m-1, n))])
    r = r - 1

    @jax.jit
    def obj(V):
        A = V[:(m-1)*r].reshape(m-1, r)
        B = V[(m-1)*r:].reshape(r, n)
        AB = jp.vstack([np.ones((1,n)), A@B])
        return jp.linalg.norm(AB - P)

    @jax.jit
    def consistency_max(V):
        A = V[:(m-1)*r].reshape(m-1, r)
        B = V[(m-1)*r:].reshape(r, n)
        AB = jp.vstack([np.ones((1,n)), A@B])
        return -(AB).flatten() + 1

    V = np.random.randn((m-1)*r + r*n)
    result = sc.optimize.minimize(obj, V,\
                                  jac=jax.jit(jax.jacrev(obj)),\
                                  tol=1e-16,\
                                  constraints=[{"type": "ineq",\
                                                "fun": consistency_max,\
                                                "jac": jax.jit(jax.jacrev(consistency_max))}],\
                                  options={"maxiter": 5000},
                                  method="SLSQP")
    A = result.x[:(m-1)*r].reshape(m-1, r)
    B = result.x[(m-1)*r:].reshape(r, n)
    AB = np.vstack([np.ones((1,n)), A@B])
    if not (np.all(AB >= 0) and np.all(AB <= 1)):
        return rand_probs_table(m, n, r+1)
    else:
        return AB

def rand_quantum_probs_table(d, m, n, r=1, field="complex"):
    effects = [np.eye(d)] + [rand_effect(d, r=r, field=field) for _ in range(m-1)]
    states = [rand_dm(d, r=r, field=field) for _ in range(n)]
    return np.array([[(e @ s).trace() for s in states] for e in effects]).real

def sample_convex_hull(hull, n):
    # https://stackoverflow.com/questions/59073952/how-to-get-uniformly-distributed-points-in-convex-hull
    points = hull.points
    dims = points.shape[-1]
    hull = points[hull.vertices]
    deln = hull[sc.spatial.Delaunay(hull).simplices]

    vols = np.abs(np.linalg.det(deln[:, :dims, :] - deln[:, dims:, :])) / np.math.factorial(dims)    
    sample = np.random.choice(len(vols), size = n, p = vols / vols.sum())

    return np.einsum('ijk, ij -> ik', deln[sample], sc.stats.dirichlet.rvs([1]*(dims + 1), size = n))


import scipy as sc
import numpy as np
from functools import partial

import jax
import jax.numpy as jp
from jax.config import config
config.update('jax_platform_name', 'cpu')
config.update("jax_enable_x64", True)

from ..utils import *
from ..povm_utils import *

@partial(jax.jit, static_argnums=(1))
def jit_spectral_inverse(P, r):
    n = P.shape[0]
    a, A, B = [1], [P], []
    for i in range(1, n+1):
        a.append(A[-1].trace()/i)
        B.append(A[-1] - a[-1]*jp.eye(n))
        A.append(P @ B[-1])
    j = n - r
    return sum([((-1 if i == 0 else 1)*a[n-j-1]*a[i]/a[n-j]**2 + \
                 (i if i < 2 else -1)*a[i-1]/a[n-j])*\
                    jp.linalg.matrix_power(P, n-j-i)
                        for i in range(r)])

@partial(jax.jit, static_argnums=(1))
def jit_pnorm(A, p):
    n = A.shape[0]
    S = jp.linalg.svd(np.eye(n) - A, compute_uv=False)
    return jp.sum(S**p)**(1/p) if p != jp.inf else jp.max(S)

def decode_norm(norm):
    norm = norm if type(norm) != None else "p2"
    if type(norm) == str:
        if norm[0] == "p":
            p = int(norm[1:])
            return jax.jit(lambda A: jit_pnorm(A, p))
    return norm

def min_quantumness(d, n=None, field="complex",
                               norm="p2",\
                               rank1=True,\
                               parallel=True,\
                               method="SLSQP",\
                               tol=1e-26,\
                               options={"disp": False,\
                                        "maxiter": 10000},\
                               max_iter=100,\
                               return_params=False):
    r = int(d*(d+1)/2) if field == "real" else int(d**2)
    n = r if type(n) == type(None) else n
    norm_func = decode_norm(norm)

    if rank1:
        if field == "complex":
            decode_params = jax.jit(lambda V: [(V[:d*n] + 1j*V[d*n:]).reshape(d, n)]*2) if parallel else\
                            jax.jit(lambda V: [(V[:d*n] + 1j*V[d*n:2*d*n]).reshape(d, n),\
                                               (V[2*d*n:3*d*n] + 1j*V[3*d*n:]).reshape(d, n)])
            initial_params = (lambda: np.random.randn(2*d*n)) if parallel else (lambda: np.random.randn(4*d*n)) 
        elif field == "real":
            decode_params = jax.jit(lambda V: [V.reshape(d, n)]*2) if parallel else\
                            jax.jit(lambda V: [V[:d*n].reshape(d, n), V[d*n:].reshape(d, n)])
            initial_params = (lambda: np.random.randn(d*n)) if parallel else (lambda: np.random.randn(2*d*n))

        def final_decode(V):
            if parallel:
                return frame_povm(np.array(decode_params(V)[0]))
            else:
                R, S = decode_params(V)
                S = jp.tile(1/jp.linalg.norm(S, axis=0), (d, 1))*S
                return frame_povm(np.array(R)), frame_povm(np.array(S))

        @jax.jit
        def wrapped_quantumness(V):
            R, S = decode_params(V)
            S = jp.tile(1/jp.linalg.norm(S, axis=0), (d, 1))*S
            P = jp.abs(R.conj().T @ S)**2
            return norm_func(jit_spectral_inverse(P, r))

        @jax.jit
        def wrapped_tightness(V):
            R, S = decode_params(V)
            return jp.linalg.norm((R @ R.conj().T) - jp.eye(d))**2
    else:
        if field == "complex":
            decode_params = jax.jit(lambda V: [(V[:n*d**2] + 1j*V[n*d**2:]).reshape(n, d, d)]*2) if parallel else\
                            jax.jit(lambda V: [(V[:n*d**2] + 1j*V[n*d**2:2*n*d**2]).reshape(n, d, d),\
                                               (V[2*n*d**2:3*n*d**2] + 1j*V[3*n*d**2:]).reshape(n, d, d)])
            initial_params = (lambda: np.random.randn(2*n*d**2)) if parallel else (lambda: np.random.randn(4*n*d**2)) 
        elif field == "real":
            decode_params = jax.jit(lambda V: [V.reshape(n, d, d)]*2) if parallel else\
                            jax.jit(lambda V: [V[:n*d**2].reshape(n, d, d), V[n*d**2:].reshape(n, d, d)]) 
            initial_params = (lambda: np.random.randn(n*d**2)) if parallel else (lambda: np.random.randn(2*n*d**2)) 

        def final_decode(V):
            make = lambda K: np.array([k.conj().T @ k for k in K])
            if parallel:
                return make(decode_params(V)[0])
            else:
                R, S = [make(x) for x in decode_params(V)]
                return R, np.array([s/s.trace() for s in S])

        @jax.jit
        def wrapped_quantumness(V):
            KR, KS = decode_params(V)
            P = jp.einsum("aji, ajk, blk, bli -> ab", KR.conj(), KR, KS.conj(), KS)/jp.tile(jp.einsum("aji, aji -> a", KS.conj(), KS), (n,1))
            return norm_func(jit_spectral_inverse(P, r))
        
        @jax.jit
        def wrapped_tightness(V):
            KR, KS = decode_params(V)
            KR = KR.reshape(d*n, d)
            return jp.linalg.norm((KR.conj().T @ KR) - jp.eye(d))**2

    for t in range(max_iter):
        result = sc.optimize.minimize(\
                    wrapped_quantumness, initial_params(),\
                    jac=jax.jit(jax.jacrev(wrapped_quantumness)),\
                    constraints=[{"type": "eq", 
                                "fun": wrapped_tightness,
                                "jac": jax.jit(jax.jacrev(wrapped_tightness))}],\
                    tol=tol,\
                    options=options,\
                    method=method)
        if not np.isclose(result.fun, float("nan"), equal_nan=True):
            if return_params:
                X = [np.array(x) for x in decode_params(result.x)]
                return X[0] if parallel else X
            return final_decode(result.x)
                


import numpy as np
import scipy as sc

np.set_printoptions(precision=3, suppress=True)

from .utils import *
from .povm_utils import *
from .random import *
from .weyl_heisenberg import *
from .sics import *
from .rics import *
from .operators import *
from .gpt import *
from .optimize import *
from .qplex import *
from .spin import *
from .grassmannian import *